@@ -1,7 +1,7 @@
 ;;;
 ;;; common-macros.scm - common macros
 ;;;
-;;;   Copyright (c) 2000-2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2000-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -180,6 +180,11 @@
     [(if-let1 var exp then . else)
      (let ((var exp)) (if var then . else))]))
 
+(define-syntax and-let1                 ;returns #f if test evaluates to #f
+  (syntax-rules ()
+    [(and-let1 var test exp exp2 ...)
+     (let ((var test)) (and var (begin exp exp2 ...)))]))
+
 (define-syntax let/cc                   ;as in PLT
   (syntax-rules ()
     [(let/cc var . body)
@@ -339,16 +344,16 @@
     [(_ (var n) . body)
      (do ([limit n]
           [var 0 (+ var 1)])
-         [(>= var limit)]
+         [(>= var limit) (undefined)]
        . body)]
     [(_ (n) . body)
      (let1 i n
-       (cond [(<= i 0)]
+       (cond [(<= i 0) (undefined)]
              [(infinite? i)
               (do () (#f) . body)] ;avoid unnecessary flonum calculation
              [else
               (do ([i i (- i 1)])
-                  [(<= i 0)]
+                  [(<= i 0) (undefined)]
                 . body)]))]
     [(_ . other)
      (syntax-error "malformed dotimes" (dotimes . other))]))
@@ -356,13 +361,16 @@
 (define-syntax dolist
   (syntax-rules ()
     [(_ (var lis res) . body)
-     (begin (for-each (lambda (var) . body) lis)
-            (let ((var '())) res))      ;bound var for CL compatibility
-     ]
+     (do ([p lis (cdr p)])
+         [(null? p)
+          (let1 var '() res)]      ;bound var for CL compatibility
+       (let1 var (car p) . body))]
     [(_ (var lis) . body)
-     (begin (for-each (lambda (var) . body) lis) '())]
+     (do ([p lis (cdr p)])
+         [(null? p) '()]
+       (let1 var (car p) . body))]
     [(_ (lis) . body)
-     (begin (for-each (lambda (_) . body) lis) '())]
+     (dolist (tmp lis) . body)]
     [(_ . other)
      (syntax-error "malformed dolist" (dolist . other))]))
 
@@ -529,3 +537,14 @@
             [r (cond-list . rest)])
        (if tmp (cons (begin . expr) r) r))]
     ))
+
+;;; ------------------------------------------------------------
+;;; er-macro-transformer
+;;;
+
+;; This will become compiled into core after 0.9.5 release; for 0.9.4,
+;; we can't compile-in define-syntaxed macros, so we keep this autoload.
+(define-syntax er-macro-transformer
+  (primitive-macro-transformer
+   (^[form def-env use-env]
+     ((with-module gauche.internal %expand-er-transformer) form use-env))))
