@@ -1,7 +1,7 @@
 ;;;
 ;;; css.scm - Generate and parse CSS
 ;;;
-;;;   Copyright (c) 2014  Shiro Kawai  <shiro@acm.org>
+;;;   Copyright (c) 2014-2016  Shiro Kawai  <shiro@acm.org>
 ;;;
 ;;;   Redistribution and use in source and binary forms, with or without
 ;;;   modification, are permitted provided that the following conditions
@@ -39,21 +39,23 @@
   (use parser.peg)
   (use util.match)
   (use text.tree)
+  (use srfi-1)
   (use srfi-13)
   (export construct-css simple-selector?
-          css-parse-file
-          css-parse-selector-string)
+          parse-css parse-css-file
+          parse-css-selector-string)
   )
 (select-module www.css)
 
 ;;;
 ;;; S-expression CSS
 ;;;
-;;;  <s-css>      : {<ruleset> | <at-rule>} ...
+;;;  <s-css>      : {<style-rule> | <at-rule>} ...
+;;;
 ;;;  <style-rule> : (style-rule <pattern> <declaration> ...)
 ;;;               | (style-decls <declaration> ...)
 ;;;
-;;;  <pattern>   : <selector> | (or <selector> ...)
+;;;  <pattern>   : <selector> | (:or <selector> ...)
 ;;;  <seletor>   : <simple-selector>
 ;;;              | <chained-selector>
 ;;;  <chained-selector> : (<simple-selector> . (<op>? . <chained-selector>))
@@ -81,7 +83,7 @@
 ;;;  <important> : !important
 ;;;  <expr>      : <term>
 ;;;              | (/ <term> <term> ...)
-;;;              | (or <term> <term> ...)
+;;;              | (:or <term> <term> ...)
 ;;;              | #(<term> <term> ...)             ; juxtaposition
 ;;;  <term>      : <quantity> | (- <quantity>) | (+ <quantity>)
 ;;;              | <string> | <ident> | <url> | <hexcolor> | <function>
@@ -93,6 +95,10 @@
 ;;;  <function>  | (<fn> <arg> ...)
 ;;;  <arg>       | <term> | #(<term> ...) | (/ <term> <term> ...)
 ;;;
+;;;  <at-rule>    : <at-media-rule> | <at-import-rule>
+;;;                     ; NB: Other at-rules are not supported yet
+;;;  <at-media-rule>  : (@media (<symbol> ...) <style-rule> ...)
+;;;  <at-import-rule> : (@import <string> (<symbol> ...))
 ;;;
 ;;; NB: style-decls shouldn't appear in a complete stylesheet, but can
 ;;;     appear when the 'style' attribute of the document is parsed.
@@ -137,7 +143,7 @@
       "}\n"))
   (define (render-pattern pattern)
     (match pattern
-      [('or selector ...) (intersperse "," (map render-selector selector))]
+      [(':or selector ...) (intersperse "," (map render-selector selector))]
       [_ (render-selector pattern)]))
   (define (render-selector selector)
     (if-let1 s (simple-selector? selector)
@@ -192,7 +198,7 @@
   (define (render-expr expr)
     (match expr
       [('/ term ...)  (intersperse "/" (map render-term term))]
-      [('or term ...) (intersperse "," (map render-expr term))]
+      [(':or term ...) (intersperse "," (map render-expr term))]
       [#(term ...) (intersperse " " (map render-expr term))]
       [term (render-term term)]))
   (define (render-term term)
@@ -445,15 +451,25 @@
                  %preserved-token)
              %WS*)))
 
-(define (block-parser opener closer tag)
+(define %component-value+  ; whitespace preserving
+  ($lazy ($/ ($seq ($tok 'WHITESPACE) ($return '(WHITESPACE)))
+             %brace-block %paren-block %bracket-block %function-call
+             %preserved-token)))
+
+(define (block-parser opener closer tag :optional (preserve-ws? #f))
   ($lift (^[_ _ v _] `(,tag ,@v))
          ($tok opener) %WS*
-         ($many ($seq ($not ($tok closer)) %component-value))
+         ($many ($seq ($not ($tok closer))
+                      (if preserve-ws? %component-value+ %component-value)))
          ($tok closer)))
 
 (define %brace-block (block-parser 'OPEN-BRACE 'CLOSE-BRACE 'brace-block))
 (define %paren-block (block-parser 'OPEN-PAREN 'CLOSE-PAREN 'paren-block))
 (define %bracket-block (block-parser 'OPEN-BRACKET 'CLOSE-BRACKET 'bracket-block))
+;; like %brace-block but we preserve whitespaces at the toplevel inside
+;; the block.  necessary to handle @media directive.
+(define %brace-block+ (block-parser 'OPEN-BRACE 'CLOSE-BRACE 'brace-block #t))
+
 (define %function-call
   ($lift (^[fn args _] `(funcall ,fn ,@args))
          ($tok 'FUNCTION)
@@ -466,7 +482,7 @@
          ($many ($seq ($not ($tok 'OPEN-BRACE))
                       ($not ($tok 'SEMICOLON))
                       %component-value))
-         ($/ %brace-block ($tok 'SEMICOLON))))
+         ($/ %brace-block+ ($tok 'SEMICOLON))))
 
 (define %qualified-rule
   ;; NB: We preserve whitespaces between qualifiers, for it is
@@ -619,7 +635,7 @@
 (define %selector-group  ;; section 5
   ($lift (^[sels] (cond [(null? sels) #f]
                         [(null? (cdr sels)) (car sels)]
-                        [else (cons 'or sels)]))
+                        [else (cons ':or sels)]))
          ($sep-by ($lift (^[_ seq _] (if (null? (cdr seq)) (car seq) seq))
                          %WS* %selector-seq %WS*)
                   ($tok 'COMMA))))
@@ -676,7 +692,7 @@
   ($lift (^[vs] (match vs [(x) x] [xs (list->vector xs)]))
          ($many ($/ ($lift (^[a _ b] `(/ ,a ,b))
                            %term ($delim #\/) %term)
-                    ($lift (^[v vs] (if (null? vs) v (cons v vs)))
+                    ($lift (^[v vs] (if (null? vs) v (cons* ':or v vs)))
                            %term ($many ($seq ($tok 'COMMA) %term)))))))
 
 (define *juxta-properties*
@@ -688,7 +704,7 @@
   (if (memq property *juxta-properties*)
     %juxtaposed-exprs
     ($do [exprs %comma-separated-exprs]
-         ($return (match exprs [(x) x] [(xs ...) `(or ,@xs)])))))
+         ($return (match exprs [(x) x] [(xs ...) `(:or ,@xs)])))))
 
 ;; val :: (property-name . tokens)
 (define (parse-declaration val)
@@ -729,40 +745,84 @@
                            block-token))
         #f))))
 
+;; Parse %at-rule.  This must be customizable.  For now, we only handle
+;; @media.
+(define (parse-at-rule at-keyword args block-token
+                       qualified-rule-handler)
+  (or (and-let* ([ (eq? at-keyword 'media) ]
+                 [syms (map (^a (match a [('IDENT . x) x] [_ #f])) args)]
+                 [tokens (match block-token [('brace-block . xs) xs] [_ #f])])
+        (let loop ([tokens tokens]
+                   [decls '()])
+          (if (null? tokens)
+            `(@media ,syms
+                     ,@(map (^p (qualified-rule-handler (car p) (cdr p)))
+                            (reverse decls)))
+            (receive (sels rest)
+                (break (^t (match t [('brace-block . _) #t] [_ #f])) tokens)
+              (if (null? rest)
+                (begin
+                  (unless (every (^s (match s [('WHITESPACE) #t][_ #f])) sels)
+                    (css-parser-warn "A selector in media block missing rules: ~s\n"
+                                     sels))
+                  (loop '() decls))
+                (loop (cdr rest) (acons sels (car rest) decls)))))))
+      (and (eq? at-keyword 'import)
+           (match args
+             [(url . media)
+              (and-let* ([url-string (match url
+                                       [('STRING . s) s]
+                                       [('URL . s) s]
+                                       [_ #f])]
+                         [media-kinds (filter-map (^x (match x
+                                                        [('IDENT . s) s]
+                                                        [_ #f]))
+                                                  media)])
+                `(@import ,url-string ,media-kinds))]
+             [_ #f]))
+      (begin
+        (css-parser-warn "Ignored unsupported at-rule: ~s\n"
+                         (list at-keyword args block-token))
+        #f)))
+
 ;; Integrated stylesheet parser
-(define (make-stylesheet-parser qualified-rule-handler)
+(define (make-stylesheet-parser qualified-rule-handler at-rule-handler)
   ($seq %inter-rule-spaces
         ($many ($followed-by
-                ($/ ($lift (^[a]
-                             (css-parser-warn "Ignored unsupported at-rule: ~s\n"
-                                              a)
-                             #f)
+                ($/ ($lift (^[a] (at-rule-handler (car a) (cadr a) (caddr a)
+                                                  qualified-rule-handler))
                            %at-rule)
                     ($lift (^[q] (qualified-rule-handler (car q) (cadr q)))
                            %qualified-rule))
                 %inter-rule-spaces))))
 
 (define (parse-stylesheet tokens
-                          :key (qualified-rule-handler parse-style-rule))
+                          :key (qualified-rule-handler parse-style-rule)
+                               (at-rule-handler parse-at-rule))
   (receive (r rest)
-      (peg-run-parser (make-stylesheet-parser qualified-rule-handler) tokens)
+      (peg-run-parser (make-stylesheet-parser qualified-rule-handler
+                                              at-rule-handler)
+                      tokens)
     (filter identity r)))
 
 ;; API
 ;; TODO: Handle @charset
-(define (css-parse-file file :key (encoding #f))
-  (call-with-input-file file
-    (^p ($ parse-stylesheet
-           $ css-tokenize $ generator->lseq $ port->char-generator p))
-    :encoding (or encoding (gauche-character-encoding))))
+(define (parse-css-file file :key (encoding #f))
+  (call-with-input-file file parse-css
+                        :encoding (or encoding (gauche-character-encoding))))
+
+;; API
+(define (parse-css :optional (port (current-input-port)))
+  ($ parse-stylesheet
+     $ css-tokenize $ generator->lseq $ port->char-generator port))
 
 ;; API
 ;; A utility function to parse only selector part.
 ;; We can't simply use parse-selectors, for it assumes the tokens is
 ;; processed as %qualified-rule.
-;; Returns #f if input is unparsable.  It is ok that imput has extra
+;; Returns #f if input is unparsable.  It is ok that input has extra
 ;; blocks; they're ignored.
-(define (css-parse-selector-string s)
+(define (parse-css-selector-string s)
   (receive (r v s) ($ %selector-only $ css-tokenize $ x->lseq s)
     (and (parse-success? r)
          (parse-selectors v))))
